@@ -3,29 +3,30 @@ import os
 import time, threading
 import re
 import inspect
+import skitai
 
-class Decorators:    
-    def __init__ (self):        
-        self.handlers = {}        
+class Decorators:
+    def __init__ (self):
+        self.handlers = {}
         self._ws_channels = {}
-        self._maintern_funcs = {}        
-        self._mount_option = {}        
+        self._maintern_funcs = {}
+        self._mount_option = {}
         self._decos = {
             "bearer_handler": self.default_bearer_handler
         }
         self._reloading = False
         self._function_specs = {}
-        self._current_function_specs = {}  
+        self._current_function_specs = {}
         self._parameter_caches = {}
-        self._conditions = {}        
-        self._cond_check_lock = threading.RLock ()                
+        self._conditions = {}
+        self._cond_check_lock = threading.RLock ()
         self._binds_server = [None] * 7
-        self._binds_request = [None] * 4        
+        self._binds_request = [None] * 4
 
     # function param saver ------------------------------------------
     def get_func_id (self,  func):
         return ("ns" in self._mount_option and self._mount_option ["ns"] + "." or "") + func.__name__
-    
+
     def save_function_spec (self, func):
         # save original function spec for preventing distortion by decorating wrapper
         # all wrapper has *args and **karg but we want to keep original function spec for auto parametering call
@@ -33,67 +34,67 @@ class Decorators:
         if func_id not in self._function_specs or func_id not in self._current_function_specs:
             # save origin spec
             self._function_specs [func_id] = inspect.getfullargspec (func)
-            self._current_function_specs [func_id] = None            
-    
+            self._current_function_specs [func_id] = None
+
     def get_function_spec (self, func):
-        # called by websocet_handler 
-        func_id = self.get_func_id (func)        
+        # called by websocet_handler
+        func_id = self.get_func_id (func)
         return self._function_specs.get (func_id)
-        
-    # app life cycling -------------------------------------------    
+
+    # app life cycling -------------------------------------------
     def before_mount (self, f):
-        self._binds_server [0] = f        
+        self._binds_server [0] = f
         return f
     start_up = before_mount
     startup = before_mount
-     
+
     def mounted (self, f):
         self._binds_server [3] = f
         return f
-    
+
     def mounted_or_reloaded (self, f):
         self._binds_server [6] = f
         return f
-    
+
     def before_reload (self, f):
         self._binds_server [5] = f
-        return f    
+        return f
     onreload = before_reload
     reload = before_reload
-    
+
     def reloaded (self, f):
         self._binds_server [1] = f
         return f
-    
+
     def before_umount (self, f):
         self._binds_server [4] = f
         return f
-    
+
     def umounted (self, f):
         self._binds_server [2] = f
         return f
     shutdown = umounted
-    
-    # Request chains ----------------------------------------------                
+
+    # Request chains ----------------------------------------------
     def before_request (self, f):
         self._binds_request [0] = f
         return f
-    
+
     def finish_request (self, f):
         self._binds_request [1] = f
         return f
-    
+
     def failed_request (self, f):
         self._binds_request [2] = f
         return f
-    
+
     def teardown_request (self, f):
         self._binds_request [3] = f
         return f
-    
+
     def testpass_required (self, testfunc):
         def decorator(f):
-            self.save_function_spec (f)    
+            self.save_function_spec (f)
             self.set_auth_flag (f, ('testpass', testfunc.__name__))
             @wraps(f)
             def wrapper (was, *args, **kwargs):
@@ -105,32 +106,38 @@ class Decorators:
                 return f (was, *args, **kwargs)
             return wrapper
         return decorator
-    
-    # parameter helpers ------------------------------------------------   
+
+    # parameter helpers ------------------------------------------------
     RX_EMAIL = re.compile (r"[a-z0-9][-.a-z0-9]*@[-a-z0-9]+\.[-.a-z0-9]{2,}[a-z]$", re.I)
     RX_UUID = re.compile (r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", re.I)
-    def _validate_param (self, params, required, ints, floats, emails, uuids, **kargs):
+    def _validate_param (self, params, required, ints, floats, emails, uuids, nones, **kargs):
         params = params or {}
         if required:
             for each in required:
                 if not params.get (each):
-                    return 'parameter {} required'.format (each)                                                
+                    return 'parameter {} required'.format (each)
         if ints:
             for each in ints:
-                try: 
+                try:
                     params [each] = int (params [each])
                 except KeyError:
-                    pass    
+                    pass
                 except ValueError:
-                    return 'parameter {} should be integer'.format (each)                                   
+                    if not params [each]:
+                        params [each] = 0
+                    else:
+                        return 'parameter {} should be integer'.format (each)
         if floats:
             for each in floats:
-                try: 
+                try:
                     params [each] = float (params [each])
                 except KeyError:
-                    pass    
+                    pass
                 except ValueError:
-                    return 'parameter {} should be float'.format (each)                    
+                    if not params [each]:
+                        params [each] = 0.0
+                    else:
+                        return 'parameter {} should be float'.format (each)
         if emails:
             for fd in emails:
                 kargs [fd] = self.RX_EMAIL
@@ -152,14 +159,14 @@ class Decorators:
                 elif val != cond:
                     return 'parameter {} is invalid'.format (fd)
                 continue
-            
+
             if len (ops) == 3:
                 if ops [1] == "len":
                     val = len (val)
                     fd = "length of {}".format (fd)
-                
+
             op = ops [-1]
-            val = (isinstance (cond, (list, tuple)) and type (cond [0]) or type (cond)) (val)                
+            val = (isinstance (cond, (list, tuple)) and type (cond [0]) or type (cond)) (val)
             if op == "neq":
                 if val == cond:
                     return 'parameter {} should be {}'.format (fd, cond)
@@ -187,10 +194,18 @@ class Decorators:
             else:
                 raise ValueError ("Unknown operator: {}".format (op))
 
-    def get_parameter_requirements (self, func_id):
-        return self._parameter_caches.get (func_id, {})        
+        if nones:
+            for each in nones:
+                try:
+                    if not params [each]:
+                        params [each] = None
+                except KeyError:
+                    pass
 
-    def require (self, scope, required = None, ints = None, floats = None, uuids = None, emails = None, **kargs):
+    def get_parameter_requirements (self, func_id):
+        return self._parameter_caches.get (func_id, {})
+
+    def require (self, scope, required = None, ints = None, floats = None, uuids = None, emails = None, nones = None, **kargs):
         def decorator(f):
             self.save_function_spec (f)
             func_id = self.get_func_id (f)
@@ -209,7 +224,7 @@ class Decorators:
                 elif scope not in ("URL", "ARGS"):
                     if was.request.method != scope:
                         return f (was, *args, **kwargs)
-                    if scope in {"GET", "DELETE"}:                    
+                    if scope in {"GET", "DELETE"}:
                         scope_ = "URL"
                     elif scope in {"POST", "PUT", "PATCH"}:
                         if was.request.get_header ("content-type", '').startswith ("application/json"):
@@ -219,15 +234,29 @@ class Decorators:
                     else:
                         return f (was, *args, **kwargs)
 
-                more_info = self._validate_param (getattr (was.request, scope_), required, ints, floats, emails, uuids, **kargs)
+                validatable_args = getattr (was.request, scope_)
+                more_info = self._validate_param (validatable_args, required, ints, floats, emails, uuids, nones, **kargs)
                 if more_info:
-                    return was.response.adaptive_error ("400 Bad Request", 'missing or bad parameterin {}'.format (scope_), 40050, more_info) 
+                    return was.response.adaptive_error ("400 Bad Request", 'missing or bad parameterin {}'.format (scope_), 40050, more_info)
+
+                # syncing args --------------------------------------
+                for k, v in validatable_args.items ():
+                    if k in kwargs:
+                        kwargs [k] = v
+                    if scope_ in ("FORM", "JSON", "URL", "DATA"):
+                        was.request.ARGS [k] = v
+                    if scope_ == "ARGS":
+                        for other_ in ("DATA", "URL"):
+                            target_ = getattr (was.request, other_)
+                            if target_ and k in target_:
+                                target_ [k] = v
                 return f (was, *args, **kwargs)
+
             return wrapper
         return decorator
     parameters_required = params_required = test_params = require
-    
-    # Automation ------------------------------------------------------    
+
+    # Automation ------------------------------------------------------
     def run_before (self, *funcs):
         def decorator(f):
             self.save_function_spec (f)
@@ -240,10 +269,10 @@ class Decorators:
                 return f (was, *args, **kwargs)
             return wrapper
         return decorator
-    
+
     def run_after (self, *funcs):
         def decorator(f):
-            self.save_function_spec (f)            
+            self.save_function_spec (f)
             @wraps(f)
             def wrapper (was, *args, **kwargs):
                 response = f (was, *args, **kwargs)
@@ -252,13 +281,13 @@ class Decorators:
                 return response
             return wrapper
         return decorator
-    
-    # Conditional Automation ------------------------------------------------------    
+
+    # Conditional Automation ------------------------------------------------------
     def _check_condition (self, was, key, func, interval, mtime_func):
         now = time.time ()
         with self._cond_check_lock:
             oldmtime, last_check = self._conditions [key]
-        
+
         if not interval or not oldmtime or now - last_check > interval:
             mtime = mtime_func (key)
             if mtime > oldmtime:
@@ -267,11 +296,11 @@ class Decorators:
                     self._conditions [key] = [mtime, now]
                 if response is not None:
                     return response
-                    
+
             elif interval:
                 with self._cond_check_lock:
-                    self._conditions [key][1] = now                        
-        
+                    self._conditions [key][1] = now
+
     def if_updated (self, key, func, interval = 1):
         def decorator(f):
             self.save_function_spec (f)
@@ -284,14 +313,14 @@ class Decorators:
                 return f (was, *args, **kwargs)
             return wrapper
         return decorator
-        
+
     def if_file_modified (self, path, func, interval = 1):
         def decorator(f):
             self.save_function_spec (f)
-            self._conditions [path] = [0, 0]            
+            self._conditions [path] = [0, 0]
             @wraps(f)
             def wrapper (was, *args, **kwargs):
-                def _getmtime (path): 
+                def _getmtime (path):
                     return os.path.getmtime (path)
                 response = self._check_condition (was, path, func, interval, _getmtime)
                 if response is not None:
@@ -299,9 +328,13 @@ class Decorators:
                 return f (was, *args, **kwargs)
             return wrapper
         return decorator
-    
+
     # Websocket ------------------------------------------------------
     def websocket (self, spec, timeout = 60, onopen = None, onclose = None, encoding = "text"):
+        use_session = False
+        if spec & skitai.WS_SESSION == skitai.WS_SESSION:
+            use_session = True
+            assert not onopen and not onclose, 'skitai.WS_SESSION cannot have onopen or onclose handler'
         def decorator(f):
             self.save_function_spec (f)
             @wraps(f)
@@ -309,39 +342,40 @@ class Decorators:
                 if not was.wshasevent ():
                     return f (was, *args, **kwargs)
                 if was.wsinit ():
-                    return was.wsconfig (spec, timeout, encoding)
+                    session = use_session and f (was, *args, **kwargs) or None
+                    return was.wsconfig (spec, timeout, encoding, session)
                 elif was.wsopened ():
                     return onopen and onopen (was) or ''
-                elif was.wsclosed ():                    
+                elif was.wsclosed ():
                     return onclose and onclose (was) or ''
             return wrapper
         return decorator
     websocket_config = websocket
-    
+
     def register_websocket (self, client_id, send):
         self._ws_channels [client_id] = send
-    
+
     def remove_websocket (self, client_id):
         try: self._ws_channels [client_id]
         except KeyError: pass
-    
+
     def websocket_send (self, client_id, msg):
-        try: 
+        try:
             self._ws_channels [client_id] (msg)
-        except KeyError: 
+        except KeyError:
             pass
-    
+
     # Mainterinancing -------------------------------------------------------
     def maintain (self, f):
         if not self._started:
             assert f.__name__ not in self._maintern_funcs, "maintain func {} is already exists".format (f.__name__)
         self._maintern_funcs [f.__name__] = f
         return f
-    
-    # Error handling ------------------------------------------------------    
+
+    # Error handling ------------------------------------------------------
     def add_error_handler (self, errcode, f, **k):
         self.handlers [errcode] = (f, k)
-        
+
     def error_handler (self, errcode, **k):
         def decorator(f):
             self.add_error_handler (errcode, f, **k)
@@ -350,10 +384,9 @@ class Decorators:
                 return f (*args, **kwargs)
             return wrapper
         return decorator
-    
+
     def default_error_handler (self, f):
         self.add_error_handler (0, f)
         return f
     defaulterrorhandler = default_error_handler
     errorhandler = error_handler
-    
\ No newline at end of file
