@@ -10,19 +10,20 @@ from types import ModuleType
 
 class MixIn (
         decorators.Decorators,
-        router.Router, 
+        router.Router,
         events.Events,
-        auth.Auth,        
+        auth.Auth,
         template_engine.TemplateEngine
-    ):    
-    
+    ):
+
     # core settings --------------------------------
     debug = False
     use_reloader = False
     auto_mount = True
-    expose_spec = False   
     enable_namespace = False
 
+    expose_spec = False
+
     # session and auth -----------------------------
     securekey = None
     session_timeout = None
@@ -36,8 +37,8 @@ class MixIn (
     def __init__ (self):
         decorators.Decorators.__init__ (self)
         router.Router.__init__ (self)
-        events.Events.__init__ (self)        
-        auth.Auth.__init__ (self)         
+        events.Events.__init__ (self)
+        auth.Auth.__init__ (self)
         template_engine.TemplateEngine.__init__ (self)
 
         self.module = None
@@ -46,43 +47,53 @@ class MixIn (
         self.logger = None
         self.store = Storage ()
         self.lock = threading.RLock ()
-        
+
         self.mount_p = "/"
         self.path_suffix_len = 0
         self.service_roots = []
         self.mount_params = {}
         self.reloadables = {}
         self.reloadable_objects = {}
-        self.last_reloaded = time.time ()        
+        self.last_reloaded = time.time ()
         self.init_time = time.time ()
-        
-        self._maintern_funcs = {}  
-        self._package_dirs = []           
-        self._mount_option = {}        
+
+        self._maintern_funcs = {}
+        self._package_dirs = []
+        self._mount_option = {}
         self._started = False
-        
-        self._salt = None        
-        
+
+        self._salt = None
+
         self.store ["__last_maintern"] = 0.0
-        self.store ["__maintern_count"] = 0        
-    
+        self.store ["__maintern_count"] = 0
+        self._locks = {}
+
+    def get_lock (self, name = None):
+        if name is None:
+            return self.lock
+        with self.lock:
+            lock = self._locks.get (name)
+            if lock is None:
+                lock = self._locks [name] = threading.RLock ()
+        return lock
+
     def maintern (self):
         if not self._maintern_funcs:
             return
         now = time.time ()
         if (now - self.store ["__last_maintern"]) < self.config.get ("maintain_interval", 60):
             return
-        
+
         was = the_was._get ()
         with self.lock:
             for func in self._maintern_funcs.values ():
                 func (was, now, self.store ["__maintern_count"])
         self.store ["__last_maintern"] = now
         self.store ["__maintern_count"] += 1
-            
+
     def get_resource (self, *args):
         return self.joinpath ("resources", *args)
-    
+
     def joinpath (self, *args):
         return os.path.normpath (os.path.join (self.home, *args))
 
@@ -94,53 +105,53 @@ class MixIn (
         else:
             self.mount_p = mount
         self.path_suffix_len = len (self.mount_p) - 1
-                
+
     def init (self, module, packagename = "app", mount = "/"):
-        self.module = module    
+        self.module = module
         self.packagename = packagename
         self.set_mount_point (mount)
-        
+
         if self.module:
             self.abspath = self.module.__file__
             if self.abspath [-3:] != ".py":
                 self.abspath = self.abspath [:-1]
             self.update_file_info    ()
-        
+
     def __getitem__ (self, k):
         return self.route_map [k]
-    
-    def get_file_info (self, module):        
+
+    def get_file_info (self, module):
         stat = os.stat (module.__file__)
         return stat.st_mtime, stat.st_size
-       
+
     def update_file_info (self):
         stat = os.stat (self.abspath)
         self.file_info = (stat.st_mtime, stat.st_size)
-        
-    #------------------------------------------------------    
+
+    #------------------------------------------------------
     @property
     def salt (self):
         if self._salt:
             return self._salt
         if not self.securekey:
             self._salt = None
-        else:            
+        else:
             self._salt = self.securekey.encode ("utf8")
         return self._salt
-    
+
     def set_default_session_timeout (self, timeout):
         self.session_timeout = timeout
-                 
+
     def set_devel (self, debug = True, use_reloader = True):
         self.debug = debug
         self.use_reloader = use_reloader
-    
+
     # services management ----------------------------------------------
     PACKAGE_DIRS = ["services", "extends", "decorative"]
     def add_package (self, *names):
         for name in names:
             self.PACKAGE_DIRS.append (name)
-    
+
     MOUNT_HOOKS = ["__mount__", "mount", "decorate"]
     UMOUNT_HOOKS = ["__umount__", "umount", "dettach"]
     def _mount (self, module):
@@ -149,9 +160,9 @@ class MixIn (
             if hasattr (module, hook):
                 mount_func = getattr (module, hook)
                 break
-                        
+
         if mount_func:
-            if not self.auto_mount and module not in self.mount_params:            
+            if not self.auto_mount and module not in self.mount_params:
                 return
             params = self.mount_params.get (module, {})
             if params.get ("debug_only") and not self.debug:
@@ -171,31 +182,31 @@ class MixIn (
             try:
                 mount_func (self)
                 self.log ("- {} mounted".format (module.__name__), "info")
-            finally:    
+            finally:
                 self._mount_option = {}
-                
+
         try:
             self.reloadables [module] = self.get_file_info (module)
         except FileNotFoundError:
-            del self.reloadables [module]  
+            del self.reloadables [module]
             return
 
         # find recursively
-        self.find_mountables (module)        
+        self.find_mountables (module)
 
     def _reload_objects (self, origin):
         if origin not in self.reloadable_objects:
             return
 
-        deletables = []    
+        deletables = []
         for objname, includers in self.reloadable_objects [origin].items ():
             for each in includers:
                 try:
                     attr = getattr (origin, objname)
-                except AttributeError:                    
+                except AttributeError:
                     deletables.append (objname)
-                    continue          
-                setattr (each, objname, attr)      
+                    continue
+                setattr (each, objname, attr)
 
         for objname in deletables:
             try:
@@ -214,27 +225,27 @@ class MixIn (
         for attr in dir (module):
             if attr.startswith ("__"):
                 continue
-            v = getattr (module, attr)            
+            v = getattr (module, attr)
             maybe_object = None
             mountable = False
-            
+
             if hasattr (v, "__module__"):
-                maybe_object = attr                
+                maybe_object = attr
                 try:
                     v = sys.modules [v.__module__]
                     if v == module:
                         continue
-                except KeyError:                    
-                    continue        
+                except KeyError:
+                    continue
 
             if type (v) is not ModuleType:
                 continue
-            
+
             try:
                 modpath = v.__spec__.origin
             except AttributeError:
                 continue
-  
+
             if not modpath:
                 continue
 
@@ -243,29 +254,45 @@ class MixIn (
                 continue
 
             for package_dir in self._package_dirs:
-                if modpath.startswith (package_dir):                    
-                    mountable = True                        
+                if modpath.startswith (package_dir):
+                    mountable = True
                     break
-        
+
             if mountable:
                 self._mount (v)
                 maybe_object and self._set_reloadable_object (maybe_object, v, module)
-    
+
     def add_package_dir (self, path):
         for exist in self._package_dirs:
             if exist.startswith (path):
                 return
-        self._package_dirs.append (path)                
+        self._package_dirs.append (path)
 
-    def mount_externals (self):        
-        for module in self.mount_params:            
+    def mount_explicit (self):
+        for module in self.mount_params:
             if module in self.reloadables:
                 continue
             self._mount (module)
-        
+
+    def mount_nested (self):
+        # within __mount__
+        for module in list (sys.modules.values ()):
+            if module in self.reloadables:
+                continue
+            try:
+                modpath = module.__spec__.origin
+            except AttributeError:
+                continue
+            if not modpath:
+                continue
+            for package_dir in self._package_dirs:
+                if modpath.startswith (package_dir):
+                    self._mount (module)
+                    break
+
     def mount (self, maybe_point = None, *modules, **kargs):
         if maybe_point:
-            if isinstance (maybe_point, str):                
+            if isinstance (maybe_point, str):
                 kargs ["point"] = maybe_point
             else:
                 modules = (maybe_point,) + modules
@@ -279,9 +306,9 @@ class MixIn (
             assert mount_func, "__mount__ hook doesn't exist"
             self.add_package_dir (os.path.dirname (module.__spec__.origin))
             self.mount_params [module] = (kargs)
-    mount_with = decorate_with = mount    
-    
-    def umount (self, *modules):        
+    mount_with = decorate_with = mount
+
+    def umount (self, *modules):
         for module in modules:
             umount_func = None
             for hook in self.UMOUNT_HOOKS:
@@ -289,61 +316,65 @@ class MixIn (
                     umount_func = getattr (module, hook)
                     break
             if not umount_func:
-                return                
+                return
             umount_func (self)
             self.log ("- %s umounted" % module.__name__, "info")
-         
+
     def umount_all (self):
         self.umount (*tuple (self.reloadables.keys ()))
     dettach_all = umount_all
-    
-    def maybe_reload (self):
-        with self.lock:  
-            if self._reloading or time.time () - self.last_reloaded < 1.0:
-                return
-            self._reloading = True
 
+    def _reload (self):
+        reloaded = 0
         for module in list (self.reloadables.keys ()):
             try:
                 fi = self.get_file_info (module)
             except FileNotFoundError:
                 del self.reloadables [module]
                 continue
-                
-            if self.reloadables [module] != fi:
-                self.log ("- reloading service, %s" % module.__name__, "info")
-                with self.glock:    
-                    for each in self.service_roots:
-                        # reinstead package root for path finder
-                        sys.modules [each.__name__] = each
+            if self.reloadables [module] == fi:
+                continue
+            self.log ("- reloading service, %s" % module.__name__, "info")
+            for each in self.service_roots:
+                # reinstead package root for path finder
+                sys.modules [each.__name__] = each
+            try:
+                newmodule = reload (module)
+            except:
+                self.log ("- reloading failed. see exception log, %s" % module.__name__, "fatal")
+                raise
+
+            reloaded += 1
+            self._current_function_specs = {}
+            self.umount (module)
+            del self.reloadables [module]
+            self._mount (newmodule)
+            self._reload_objects (newmodule)
+        return reloaded
+
+    def maybe_reload (self):
+        with self.lock:
+            if self._reloading or time.time () - self.last_reloaded < 1.0:
+                return
+            self._reloading = True
+
+        try:
+            self._reload () and self.load_jinja_filters ()
+        finally:
+            with self.lock:
+                self.last_reloaded = time.time ()
+                self._reloading = False
 
-                    try:
-                        newmodule = reload (module)
-                    except:
-                        self.log ("- reloading failed. see exception log, %s" % module.__name__, "fatal")
-                        raise    
-                                            
-                self._current_function_specs = {}
-                self.umount (module)
-                del self.reloadables [module]
-                self._mount (newmodule)                    
-                self._reload_objects (newmodule)
-                self.load_jinja_filters ()       
-        
-        with self.lock:  
-            self.last_reloaded = time.time ()
-            self._reloading = False
-    
     # logger ----------------------------------------------------------
     def set_logger (self, logger):
-        self.logger = logger 
-        
+        self.logger = logger
+
     def log (self, msg, type = "info"):
         self.logger (msg, type)
-    
+
     def trace (self):
         self.logger.trace ()
-   
+
     PHASES = {
         'before_mount': 0,
         'mounted': 3,
@@ -359,8 +390,8 @@ class MixIn (
         index = self.PHASES.get (phase)
         func = self._binds_server [index]
         if not func:
-            return        
-        
+            return
+
         obj.app = self
         try:
             func (obj)
@@ -368,9 +399,9 @@ class MixIn (
             if self.logger:
                 self.logger.trace ()
             else:
-                raise                             
-    
-    # Error handling ------------------------------------------------------        
+                raise
+
+    # Error handling ------------------------------------------------------
     def render_error (self, error, was = None):
         handler = self.handlers.get (error ['code'], self.handlers.get (0))
         if not handler:
@@ -383,32 +414,31 @@ class MixIn (
         content =  handler [0] (was, error)
         was.app = None
         return content
-    
 
-    # app startup and shutdown --------------------------------------------    
-    def cleanup (self):   
+
+    # app startup and shutdown --------------------------------------------
+    def cleanup (self):
         pass
-            
+
     def _start (self, wasc, route, reload = False):
         self.wasc = wasc
         if not route:
             self.basepath = "/"
-        elif not route.endswith ("/"):            
+        elif not route.endswith ("/"):
             self.basepath = route + "/"
         else:
             self.basepath = route
-        
+
     def start (self, wasc, route):
         self.bus.emit ("app:starting", wasc)
         self._start (wasc, route)
         self.bus.emit ("app:started", wasc)
         self._started = True
-        
-    def restart (self, wasc, route):        
+
+    def restart (self, wasc, route):
         self._reloading = True
-        self.bus.emit ("app:restarting", wasc)    
+        self.bus.emit ("app:restarting", wasc)
         self._start (wasc, route, True)
-        self.bus.emit ("app:restarted", wasc)    
+        self.bus.emit ("app:restarted", wasc)
         self._reloading = False
-    
-    
\ No newline at end of file
+
